from abc import ABCMeta, abstractmethod, abstractproperty

class Sigma(metaclass=ABCMeta):

    @abstractproperty
    def commit(self, m, r):
        return


    @abstractproperty
    def challenge(self, e):
        return

    @abstractproperty
    def proof(self):
        return


class Commitment(metaclass=ABCMeta):
    C = abstractproperty()
    D = abstractproperty()


class TrapdoorCommitment(Commitment):
    trapdoor = abstractproperty()



from klefki.crypto.ssss import SSSS
from klefki.types.algebra.groups import EllipticCurveGroup
from klefki.types.algebra.fields import FiniteField
from klefki.types.algebra.utils import randfield
from functools import reduce
from operator import mul


class VSS(SSSS):
    def __init__(self, G: EllipticCurveGroup, F: FiniteField, secret=None, k=None, n=None, poly_params=None):
        self.G = G
        super().__init__(F, secret, k, n, poly_params)


    def setup(self, secret, k, n, poly_params = []):
        if not poly_params:
            poly_params = [randfield(self.F) for _ in range(k - 1)]

        self.poly_proofs = [self.G ** p for p in [secret] + poly_params]

        return super().setup(secret, k, n, poly_params)


    def verify(self, x, y):
        return self.verify_proofs(x, y, self.poly_proofs, self.G, self.k)


    @staticmethod
    def verify_proofs(x, y, proofs, G, k):
        return G ** y == reduce(mul, [proofs[i] ** (x ** i) for i in range(k)])



from abc import ABCMeta, abstractmethod, abstractproperty
from klefki.algorithms import double_and_add_algorithm
from klefki.algorithms import complex_truediv_algorithm
from klefki.numbers import modular_sqrt


class Functor(metaclass=ABCMeta):

    __slots__ = ['value']

    def __init__(self, v):
        if isinstance(v, self.__class__):
            self.value = v.value
        self.value = self.fmap(v)

    def fmap(self, o):
        return o

    @property
    def functor(self):
        return self.__class__

    def id(self):
        return self.value

class Groupoid(Functor):

    __slots__ = ()

    @abstractmethod
    def op(self, g: 'Group') -> 'Group':
        pass

    def __eq__(self, b) -> bool:
        return self.value == b.value

    def __lt__(self, b) -> bool:
        return self.value < b.value

    def __le__(self, b) -> bool:
        return self.value <= b.value

    def __gt__(self, b) -> bool:
        return self.value > b.value

    def __ge__(self, b) -> bool:
        return self.value >= b.value

    def __add__(self, g: 'Group') -> 'Group':
        '''
        Allowing call associativity operator via A + B
        Strict limit arg `g` and ret `res` should be subtype of Group,
        For obeying axiom `closure` (1)
        '''
        res = self.op(g)
        assert isinstance(res, type(self))
        return res

    def __radd__(self, g):
        return self.__add__(g)

    def __mul__(self, g: 'Group') -> 'Group':
        return self.__add__(g)


    def __repr__(self):
        return "%s::%s" % (
            type(self).__name__,
            self.value
        )

    def __str__(self):
        return str(self.value)


class SemiGroup(Groupoid):

    __slots__ = ()

    @abstractmethod
    def op(self, g: 'Group') -> 'Group':
        '''
        The Operator for obeying axiom `associativity` (2)
        '''
        pass


class Monoid(SemiGroup):
    __slots__ = ()

    @abstractproperty
    def identity(self):
        '''
        The value for obeying axiom `identity` (3)
        '''
        pass

    def __not__(self):
        return self is not self.identity

    def scalar(self, times):
        while getattr(times, 'value', None) != None:
            times = times.value
        if times == 0:
            return self.identity
        return double_and_add_algorithm(times, self, self.identity)

    def __matmul__(self, times):
        return self.scalar(times)

    def __pow__(self, times) -> 'Group':
        return self.scalar(times)

    def __xor__(self, times) -> 'Group':
        return self.__matmul__(times)



class Group(Monoid):
    __slots__ = ()

    @abstractmethod
    def inverse(self: 'Group') -> 'Group':
        '''
        Implement for axiom `inverse`
        '''
        pass

    def __sub__(self, g: 'Group') -> 'Group':
        '''
        Allow to reverse op via a - b
        '''
        return self.op(g.inverse())

    def __neg__(self) -> 'Group':
        return self.inverse()


class Field(Group):
    __slots__ = ()

    @abstractmethod
    def sec_op(self, g: 'Field') -> 'Field':
        '''
        The Operator for obeying axiom `associativity` (2)
        '''
        pass

    @abstractmethod
    def sec_inverse(self) -> 'Field':
        '''
        Implement for axiom `inverse`
        '''
        pass

    @abstractmethod
    def sec_identity(self):
        pass

    def __invert__(self):
        return self.sec_inverse()

    def __mul__(self, g: 'Field') -> 'Field':
        '''
        Allowing call associativity operator via A * B
        Strict limit arg `g` and ret `res` should be subtype of Group,
        For obeying axiom `closure` (1)
        '''
        res = self.sec_op(g)
        assert isinstance(res, type(self)), 'result shuould be %s' % type(self)
        return res

    def __pow__(self, b, m=None):
        if hasattr(b, "value"):
            b = b.value

        if b == (1/2):
            root = modular_sqrt(self.value, self.P)
            assert root != 0, "ins dont have root"
            return self.__class__(root)


        if hasattr(self, "P"):
            m = self.P
            if b < 0:
                return ~self.__class__(pow(self.value, b * -1, m))
        return self.__class__(pow(self.value, b, m))
        # If b == 0:
        #     return self.__class__(1)
        # if 0 < b < 1:
        #     return self.__class__(self.value ** b)
        # if b == 1:
        #     return self
        # if b == 2:
        #     return self * self
        # return self * (self ** (b - 1))

    def __truediv__(self, g: 'Field') -> 'Field':
        if isinstance(g.value, complex):
            return complex_truediv_algorithm(complex(1), self.value, self.__class__)
        return self.sec_op(g.sec_inverse())



from klefki.types.algebra.concrete import (
    EllipticCurveGroupSecp256k1 as ECG,
    FiniteFieldCyclicSecp256k1 as CF,
    FiniteFieldSecp256k1 as F
)
from klefki.types.algebra.isomorphism import bijection
from random import randint

def encode(key: ECG):
    x = hex(key.value[0].value)[2:]
    y = hex(key.value[1].value)[2:]
    return '0' * (32 - len(x)) + x + '0' * (32 - len(y)) + y


@bijection(encode)
def decode(key: str) -> ECG:
    x = F(int(key[:32], 16))
    y = F(int(key[32:], 16))
    return ECG((x, y))


def randfield(F):
    return F(randint(0, F.P - 1))

def randgroup(G):
    return G(randint(0, G.N - 1))



#
# https://github.com/mikeivanov/paillier/blob/master/paillier/primes.py

import random
import sys

def ipow(a, b, n):
    """calculates (a**b) % n via binary exponentiation, yielding itermediate
       results as Rabin-Miller requires"""
    A = a = a % n
    yield A
    t = 1
    while t <= b:
        t <<= 1

    # t = 2**k, and t > b
    t >>= 2

    while t:
        A = (A * A) % n
        if t & b:
            A = (A * a) % n
        yield A
        t >>= 1

def rabin_miller_witness(test, possible):
    """Using Rabin-Miller witness test, will return True if possible is
       definitely not prime (composite), False if it may be prime."""
    return 1 not in ipow(test, possible - 1, possible)

smallprimes = (2,3,5,7,11,13,17,19,23,29,31,37,41,43,
               47,53,59,61,67,71,73,79,83,89,97)

def default_k(bits):
    return max(40, 2 * bits)

def is_probably_prime(possible, k=None):
    if possible == 1:
        return True
    if k is None:
        k = default_k(possible.bit_length())
    for i in smallprimes:
        if possible == i:
            return True
        if possible % i == 0:
            return False
    for i in range(k):
        test = random.randrange(2, possible - 1) | 1
        if rabin_miller_witness(test, possible):
            return False
    return True


def generate_prime(bits, k=None):
    """Will generate an integer of b bits that is probably prime
       (after k trials). Reasonably fast on current hardware for
       values of up to around 512 bits."""
    assert bits >= 8

    if k is None:
        k = default_k(bits)

    while True:
        possible = random.randrange(2 ** (bits-1) + 1, 2 ** bits) | 1
        if is_probably_prime(possible, k):
            return possible



from functools import reduce
from klefki.crypto.ecdsa.secp256k1 import (
    pubkey
)

from klefki.types.algebra.concrete import (
    EllipticCurveCyclicSubgroupSecp256k1 as CG,
    FiniteFieldSecp256k1 as F,
    FiniteFieldCyclicSecp256k1 as CF
)
from klefki.types.algebra.isomorphism import bijection


G = CG.G
A = CG.A
B = CG.B
P = F.P


def encode_pubkey(pub: CG) -> str:
    n = hex(pub.value[0].value).replace('0x', '')
    return '0' + str(2 + (pub.value[1].value % 2)) + ((64 - len(n)) * '0' + n)


@bijection(encode_pubkey)
def decode_pubkey(pub: str) -> CG:
    pub = bytearray.fromhex(pub)
    x = reduce(lambda x, y: x * 256 + y, bytes(pub[1:33]), 0)
    beta = pow(int(x * x * x + A * x + B), int((P + 1) // 4), int(P))
    y = (P - beta) if ((beta + pub[0]) % 2) else beta
    return CG(
        (
            F(x),
            F(y)
        )
    )


def gen_pub_key(key: CF) -> str:
    return encode_pubkey(
        pubkey(key)
    )



import base58
import base64
from typing import Iterable, Iterator, Callable
from klefki.types.algebra.isomorphism import bijection
from klefki.types.algebra.concrete import (
    FiniteFieldCyclicSecp256k1 as CF,
)

from hashlib import sha256
import hashlib
from functools import reduce

b58encode = bijection(base58.b58decode)(base58.b58encode)
b64encode = bijection(base64.b64decode)(base64.b64encode)


def int_to_byte(key: int, length=32) -> bytes:
    return key.to_bytes(length, byteorder='big')


@bijection(int_to_byte)
def byte_to_int(byte: bytes) -> CF:
    return int.from_bytes(byte, "big")


def dhash256(x: int) -> int:
    return sha256(sha256(x).digest()).digest()


def ripemd160(x) -> int:
    ripemd160 = hashlib.new('ripemd160')
    if isinstance(x, str):
        x = x.encode()
    if isinstance(x, list):
        for i in x:
            ripemd160.update(i)
    else:
        ripemd160.update(x)
    return ripemd160.digest()


def trunks(l: Iterable, n: int) -> Iterator:
    return zip(*[iter(l)] * n)


def concat(a: bytes, b: bytes) -> bytes:
    return bytes(a) + bytes(b)


def compose(*fs: Iterable[Callable]):
    def _(*args, **kwargs):
        return reduce(
            lambda x, y: y(x),
            fs,
            kwargs or args
        )
    return _


def to_sha256int(m: str):
    return int.from_bytes(sha256(m.encode()).digest(), 'big')


class EnumDict(dict):
    def __getattr__(self, k):
        '''
        Support Lazy
        '''
        value = self.get(k)
        if callable(value):
            value = value()
        if type(value) is dict:
            return EnumDict(value)
        else:
            return value

    def __setattr__(self, k, v):
        self.update({k: v})
        return self

    def __contains__(self, v):
        return v in self.values()


def parse_lv_format(b):
    i = 0
    j = i + 4
    ret = []
    while j <= len(b):
        l = int.from_bytes(bytes(b[i:j]), "big")
        i = j
        j = j + l
        v = bytes(b[i:j])
        ret.append(v)
        i = j
        j = i + 4
    return ret


def CF2Bytes(cf, l=32):
    return cf.value.to_bytes(l, byteorder="big")



from typing import Tuple, TypeVar, Iterable


__all__ = [
    'extended_euclidean_algorithm',
    'double_and_add_algorithm'
]

T = TypeVar('T')


def complex_truediv_algorithm(x: complex, y: complex, f: T) -> T:
    a = f(x.real)
    b = f(x.imag)
    c = f(y.real)
    d = f(y.imag)

    return f(complex(
        (a*c + b*d) / (c**2 + d**2),
        (b*c - a*d) / (c**2 + d**2)
    ))


def extended_euclidean_algorithm(a: int, b: int) -> Tuple[int, int, int]:
    '''
    Returns a three-tuple (gcd, x, y) such that
    a * x + b * y == gcd, where gcd.

    This function implements the extended Euclidean
    algotithm and runs in O(log b) in the worst case
    '''
    s, t, r = 0, 1, b
    old_s, old_t, old_r = 1, 0, a

    while r != 0:
        quoient = old_r // r
        old_r, r = r, old_r - quoient * r
        old_s, s = s, old_s - quoient * s
        old_t, t = t, old_t - quoient * t

    return old_r, old_s, old_t


def double_and_add_algorithm(times: int, x: T, init: T) -> T:
    """
    Returns the result of n * x, computed using
    the double and add algorithm.
    """
    def bits(n: int) -> Iterable:
        """
        Generates the binary digits of n, starting
        from the least significant bit.
        bits(151) -> 1, 1, 1, 0, 1, 0, 0, 1
        """
        while n:
            yield n & 1
            n >>= 1

    result = init
    addend = x

    for bit in bits(times):
        if bit == 1:
            result = addend + result
        addend = addend + addend

    return result


def newton_iterator_sqrt(x: T):
    if (x.value == 0): return x
    last = x.__class__(0)
    res = x.__class__(1)
    while res != last:
        last = res
        res = (res + x / res) / x.__class__(2)
    return res



import os
import base64
import M2Crypto


class SymmetricEncryption(object):

    def __init__(self, key, iv, alg="aes_128_cbc"):
        self.iv = iv
        self.key = key
        self.alg = alg

    def encrypt(self, plaintext):
        ENCRYPT = 1
        cipher = M2Crypto.EVP.Cipher(alg=self.alg, key=self.key, iv=self.iv, op=ENCRYPT)
        ciphertext = cipher.update(plaintext) + cipher.final()
        return base64.b64encode(ciphertext)

    def decrypt(self, cyphertext):
        DECRYPT = 0
        cipher = M2Crypto.EVP.Cipher(alg=self.alg, key=self.key, iv=self.iv, op=DECRYPT)
        plaintext = cipher.update(base64.b64decode(cyphertext)) + cipher.final()
        return plaintext



# Auto-generated by asn1ate v.0.6.0 from signature.asn
# (last modified on 2017-11-22 15:15:57)

from pyasn1.type import univ, char, namedtype, namedval, tag, constraint, useful


class ECDSA_Sig_Value(univ.Sequence):
    pass


ECDSA_Sig_Value.componentType = namedtype.NamedTypes(
    namedtype.NamedType('r', univ.Integer()),
    namedtype.NamedType('s', univ.Integer())
)





from functools import reduce
from operator import add
from klefki.types.algebra.utils import randfield


def additive_share(secret, F, n):
    shares = [randfield(F) for _ in range(n - 1)]
    shares += [(F(secret) - reduce(add, shares))]
    return shares

def additive_reconstruct(shares):
    return reduce(add, shares)



from typing import Iterable, Iterator
from itertools import starmap
from klefki.utils import dhash256, concat, trunks

__all__ = ['MerkleTree']


class MerkleTree():
    def __init__(self, data: Iterable, child=None) -> None:
        data = list(data)
        if not child and len(data) % 2 != 0:
            data.append(data[-1])
        self.data = data
        self._nodes = None
        self._parents = None

    @property
    def nodes(self) -> Iterator:
        if self._nodes is None:
            self._nodes = list(map(dhash256, self.data))
        return self._nodes

    @property
    def parents(self) -> Iterator:
        if self._parents is None:
            self._parents = self.__class__(
                map(
                    dhash256,
                    list(starmap(
                        concat,
                        trunks(self.nodes, 2)
                    ))
                ),
                child=self
            )
        return self._parents

    @property
    def root(self):
        if len(self.parents.nodes) > 1:
            return self.parents.root
        else:
            return self.parents



from klefki.crypto.ecdsa.secp256k1 import (
    random_privkey,
)
from klefki.types.algebra.concrete import (
    FiniteFieldCyclicSecp256k1 as CF,
)
from klefki.types.algebra.isomorphism import bijection

def encode_privkey(key: CF) -> str:
    return hex(key.value)[2:]


@bijection(encode_privkey)
def decode_privkey(key: str) -> CF:
    return CF(int(key, 16))


def gen_random_privkey():
    return random_privkey()



#
# Ref:
#* I. Damg˚ard, J. Groth. Non-interactive and reusable non-malleable commitment schemes. Proc. of 35 th ACM Symp. on Theory of Computing (STOC’03), pp.426- 437, 2003.
#
#
from klefki.types.algebra.concrete import EllipticCurveCyclicSubgroupSecp256k1 as ECC
from klefki.types.algebra.concrete import EllipticCurveGroupSecp256k1 as Cruve
from klefki.types.algebra.concrete import FiniteFieldCyclicSecp256k1 as CF
from klefki.types.algebra.concrete import FiniteFieldSecp256k1 as F
from klefki.types.algebra.utils import randfield
from klefki.bitcoin.address import gen_address
from klefki.utils import to_sha256int
import hmac
from klefki.bitcoin.public import encode_pubkey
from klefki.utils import int_to_byte
from klefki.zkp.commitment import Commitment, Sigma
from klefki.zkp.pedersen import PedersonCommitment

G = ECC.G
H = Cruve.lift_x(CF(to_sha256int("hello NIRNCS")))


def keygen(F):
    pk = randfield(F)
    vk = F(to_sha256int(str(pk.value)))
    h = gen_address
    return (pk, vk, h)




class NMCommitment(Commitment):

    def __init__(self, G, H, key=None):
        if not key:
            key = keygen(CF)
        self.key = key

    def commit(self, m, ak):
        pk, vk, h = self.key
        com = PedersonCommitment(H, G, pk, ak)
        # (𝑐,𝑑)=𝐻𝑆𝑐𝑜𝑚𝑚𝑖𝑡𝑝𝑘(𝑎𝑘)
        c, d = com.C, com.D
        # α=ℎ(𝑐)
        alpha = CF(to_sha256int(gen_address(c)))
        # Now we simulate a proof of knowledge of a signature on 𝛼 with challenge 𝑚
        S = PedersonCommitment(H, G, vk, alpha)
        r1, r2 = randfield(vk.functor), randfield(vk.functor)
        a = S.commit(r1, r2)
        z = S.challenge(m)
        self.S = S
        # Finally, we compute 𝑚𝑎𝑐=𝑀𝐴𝐶𝑎𝑘(𝑎).
        mac = hmac.new(str(ak.value).encode(), encode_pubkey(a).encode()).hexdigest()
        self.c = (c, a, mac)
        self.d = (m, d , z)


    @property
    def C(self):
        return self.c

    @property
    def D(self):
        return self.d


    def open(self):
        c, a, mac = self.c
        m, d, z = self.d
        alpha = CF(to_sha256int(gen_address(c)))
        assert mac == hmac.new(str(d[1].value).encode(), encode_pubkey(a).encode()).hexdigest()
        assert self.S.proof
        return m



from klefki.zkp.commitment import TrapdoorCommitment
from klefki.types.algebra.abstract import Group, Field
from klefki.types.algebra.meta import field
from functools import reduce, partial
from typing import Iterator
from operator import add

__all__ = ['commitment', 'vertex_commitment', 'com', 'PedersonCommitment']

def v_multi(g: [Group], a: [Field]) -> [Group]:
    return reduce(lambda x,y: x+y,
                  list(map(lambda a: a[0] @ a[1], zip(g, a))))


def commitment(x: Field, r: Field, H: Group, G: Group) -> Group:
    return (G ^ x) * (H ^ r)


def vertex_commitment(x: [Field], r: Field, H: Group, G: [Group]) -> Group:
    return v_multi(G, x) * H ^ r


def matrix_commitment(x: [[Field]], r: Field, H: Group, G: [Group]) -> Group:
    return vertex_commitment(G, H, reduce(add, x), r)

def com(x, r, H, G) -> Group:
    '''
    Com(x, r) = xG + rH
    '''
    if type(x) in [
            Iterator,
            list
    ]:
        if type(x[0]) in [
            Iterator,
            list
        ]:
            return matrix_commitment(x, r, H, G)
        return vertex_commitment(x, r, H, G)
    return commitment(x, r, H, G)


class PedersonCommitment(TrapdoorCommitment):


    def __init__(self, G, H, x, r):
        '''
        G, H <- ECC
        '''
        self.G = G
        self.H = H
        self.com = partial(com, G=G,H=H)
        self.x = x
        self.r = r
        self.A = self.com(self.x, self.r)


    def commit(self, y, s):
        self.B = self.com(y, s)
        self.c = self.B
        self.y = y
        self.s = s
        return self.c

    def trapdoor(self, new_secret, x):
        '''
        x is trapdoor
        '''
        r_ = self.r - (new_secret - self.x) * ~x
        self.A = self.com(new_secret, r_)

    def challenge(self, e):
        '''
        e is the random callange
        '''
        self.e = e
        self.response = (
            self.y * self.e + self.x,
            self.s * self.e + self.r
        )
        return self.response


    def proof(self, trans=None):
        if not trans:
            trans = self.transcript
        B, e, s = trans
        assert self.com(*s) == self.A * (self.B ** e)
        return True

    @property
    def transcript(self):
        return (self.c, self.e, self.response)

    @property
    def C(self):
        return self.A

    @property
    def D(self):
        return (self.x, self.r)


    @staticmethod
    def verify(H, G, C, D):
        if C == com(*D, H=H, G=G):
            return D[1]



import math
from functools import reduce
from operator import mul

# https://crypto.stackexchange.com/questions/29591/lcm-versus-phi-in-rsa

def rsa_lambda(p, q):
    return lcm(p-1, q-1)

def rsa_phi(p, q):
    return (p-1) * (q-1)


def fn_lambda(n):
    '''
    Carmichael Function
    A number n is said to be a Carmichael number if
    it satisfies the following modular arithmetic condition:
    power(b, n-1) MOD n = 1,
    for all b ranging from 1 to n such that b and
    n are relatively prime, i.e, gcd(b, n) = 1
    '''
    coprimes = [x for x in range(1, n) if math.gcd(x, n) == 1]
    k = 1
    while not all(pow(x, k, n) == 1 for x in coprimes):
        k += 1
    return k


def fn_phi(n):
    '''
    Eulers Totient Function
    https://stackoverflow.com/questions/18114138/computing-eulers-totient-function
    '''
    amount = 0
    for k in range(1, n + 1):
        if math.gcd(n, k) == 1:
            amount += 1
    return amount


carmichael = fn_lambda
totient = fn_phi

def lcm(a, b):
    return abs(a*b) // math.gcd(a, b)


def length(n, base=10):
    return math.floor(math.log(n, base)) + 1


def Blength(n, base=2):
    return math.floor(math.log(n, base)) + 1


def Dlength(n, base=10):
    return math.floor(math.log(n, base)) + 1


def power(base, exp):
    """ Fast power calculation using repeated squaring """
    if exp < 0:
        return 1 / power(base, -exp)
    ans = 1
    while exp:
        if exp & 1:
            ans *= base
        exp >>= 1
        base *= base
    return ans


def invmod(a, p):
    from klefki.types.algebra.meta import field
    f = field(p)
    return (~f(a)).value


def modpow(base, exponent, modulus):
    """Modular exponent:
         c = b ^ e mod m
       Returns c.
       (http://www.programmish.com/?p=34)"""
    result = 1
    while exponent > 0:
        if exponent & 1 == 1:
            result = (result * base) % modulus
        exponent = exponent >> 1
        base = (base * base) % modulus
    return result




def modular_sqrt(a, p):
    """ Find a quadratic residue (mod p) of 'a'. p
    must be an odd prime.
    Solve the congruence of the form:
    x^2 = a (mod p)
    And returns x. Note that p - x is also a root.
    0 is returned is no square root exists for
    these a and p.
    The Tonelli-Shanks algorithm is used (except
    for some simple cases in which the solution
    is known from an identity). This algorithm
    runs in polynomial time (unless the
    generalized Riemann hypothesis is false).
    """
    # Simple cases
    #
    if legendre_symbol(a, p) != 1:
        return 0
    elif a == 0:
        return 0
    elif p == 2:
        return p
    elif p % 4 == 3:
        return pow(a, (p + 1) // 4, p)

    # Partition p-1 to s * 2^e for an odd s (i.e.
    # reduce all the powers of 2 from p-1)
    #
    s = p - 1
    e = 0
    while s % 2 == 0:
        s /= 2
        e += 1

    # Find some 'n' with a legendre symbol n|p = -1.
    # Shouldn't take long.
    #
    n = 2
    while legendre_symbol(n, p) != -1:
        n += 1

    # Here be dragons!
    # Read the paper "Square roots from 1; 24, 51,
    # 10 to Dan Shanks" by Ezra Brown for more
    # information
    #

    # x is a guess of the square root that gets better
    # with each iteration.
    # b is the "fudge factor" - by how much we're off
    # with the guess. The invariant x^2 = ab (mod p)
    # is maintained throughout the loop.
    # g is used for successive powers of n to update
    # both a and b
    # r is the exponent - decreases with each update
    #
    x = pow(a, (s + 1) // 2, p)
    b = pow(a, s, p)
    g = pow(n, s, p)
    r = e

    while True:
        t = b
        m = 0
        for m in xrange(r):
            if t == 1:
                break
            t = pow(t, 2, p)

        if m == 0:
            return x

        gs = pow(g, 2 ** (r - m - 1), p)
        g = (gs * gs) % p
        x = (x * gs) % p
        b = (b * g) % p
        r = m


def legendre_symbol(a, p):
    """ Compute the Legendre symbol a|p using
    Euler's criterion. p is a prime, a is
    relatively prime to p (if p divides
    a, then a|p = 0)
    Returns 1 if a has a square root modulo
    p, -1 otherwise.
    """
    ls = pow(a, (p - 1) // 2, p)
    return -1 if ls == p - 1 else ls




def crt(a_list, n_list) -> int:
    """Applies the Chinese Remainder Theorem to find the unique x such that x = a_i (mod n_i) for all i.
    :param a_list: A list of integers a_i in the above equation.
    :param n_list: A list of integers b_i in the above equation.
    :return: The unique integer x such that x = a_i (mod n_i) for all i.
    copy from::
    https://github.com/cryptovoting/damgard-jurik/blob/master/damgard_jurik/utils.py#L100
    """
    a_list = [a_i for a_i in a_list]
    n_list = [n_i for n_i in n_list]

    N = reduce(mul, n_list)
    y_list = [N // n_i for n_i in n_list]
    z_list = [invmod(y_i, n_i) for y_i, n_i in zip(y_list, n_list)]
    x = sum(a_i * y_i * z_i for a_i, y_i, z_i in zip(a_list, y_list, z_list))

    return x



from sha3 import keccak_256
from klefki.crypto.ecdsa.secp256k1 import pubkey
from klefki.utils import int_to_byte
from klefki.types.algebra.concrete import (
    EllipticCurveGroupSecp256k1 as ECG,
    FiniteFieldCyclicSecp256k1 as CF,
)


def gen_address(pub: ECG) -> str:
    x = int_to_byte(pub.value[0].value)
    y = int_to_byte(pub.value[1].value)
    return gen_address_from_bytes(x + y)


def gen_address_from_bytes(pub: bytes) -> str:
    return '0x%s' % keccak_256(pub).hexdigest()[24:].upper()


def gen_address_from_priv(priv: CF) -> str:
    return gen_address(pubkey(priv))

def gen_address_from_priv_bytes(priv: bytes) -> str:
    sk = CF(int.from_bytes(priv, "big"))
    return gen_address(pubkey(sk))



from math import gcd
from abc import ABCMeta
from random import randint
from klefki.types.algebra.utils import randfield
from klefki.types.algebra.meta import field
from klefki.numbers import lcm
from functools import partial
import random


def L(x, n):
    return (x - 1) // n


class Paillier():

    def __init__(self, P, Q):
        assert gcd(P * Q, (P - 1) * (Q - 1)) == 1

        N = P * Q
        Lam = lcm(P - 1, Q - 1)

        F = field(N)
        DF = field(N**2)
        G = randfield(DF)

        M = ~F(L(pow(G, Lam).value, N))
        self.N = N
        self.G = G
        self.privkey = Lam
        self.pubkey = (self.N, self.G)


    @classmethod
    def encrypt(cls, m, pub):
        N, G = pub

        if hasattr(m, "value"):
            m = m.value

        r = G.functor(random.randint(0, N))
        return G**m * r**N


    @classmethod
    def decrypt(cls, c, priv, pub):
        Lam = priv
        N, G = pub

        F = field(N, "N")
        return F(L((c ** Lam).value, N)) * ~F(L(pow(G, Lam).value, N))

    def E(self, m, pub=None):
        return self.encrypt(m, pub or self.pubkey)


    def D(self, c, priv=None, pub=None):
        return self.decrypt(c, priv or self.privkey, pub or self.pubkey)


    def test(self):
        m = random.randint(0, self.N)
        assert m == self.D(self.E(m)).value
        return self



from operator import add
from functools import reduce
from klefki.utils import EnumDict
# Ref: https://en.bitcoin.it/wiki/Script
# Constants
CONSTS = EnumDict(dict(
    OP_0=bytes([0x00]),
    OP_FALSE=bytes([0x00]),
    OP_PUSHDATA1=bytes([0x4c]),
    OP_PUSHDATA2=bytes([0x4d]),
    OP_PUSHDATA4=bytes([0x4e]),
    OP_1NEGATE=bytes([0x4f]),
    OP_1=bytes([0x51]),
    OP_TRUE=bytes([0x51]),
    OP_2=bytes([0x52]),
    OP_3=bytes([0x53]),
    OP_4=bytes([0x54]),
    OP_5=bytes([0x55]),
    OP_6=bytes([0x56]),
    OP_7=bytes([0x57]),
    OP_8=bytes([0x58]),
    OP_9=bytes([0x59]),
    OP_10=bytes([0x5a]),
    OP_11=bytes([0x5b]),
    OP_12=bytes([0x5c]),
    OP_13=bytes([0x5d]),
    OP_14=bytes([0x5e]),
    OP_15=bytes([0x5f]),
    OP_16=bytes([0x60])
))

FLOW_CONTROL = EnumDict(dict(

    # Flow Control

    OP_NOP=bytes([0x61]),
    OP_IF=bytes([0x99]),
    OP_NOTIF=bytes([0x100]),
    OP_ELSE=bytes([0x103]),
    OP_ENDIF=bytes([0x104]),
    OP_VERIFY=bytes([0x69]),
    OP_RETURN=bytes([0x6a]),
))
# Stack

STACK = EnumDict(dict(
    OP_TOTALSTACK=bytes([0x6b]),
    OP_FORMALSTACK=bytes([0x6c]),
    OP_IFDUP=bytes([0x73]),
    OP_DEPTH=bytes([0x74]),
    OP_DROP=bytes([0x75]),
    OP_DUP=bytes([0x76]),
    OP_NIP=bytes([0x77]),
    OP_OVER=bytes([0x78]),
    OP_PICK=bytes([0x79]),
    OP_ROLL=bytes([0x7a]),
    OP_ROT=bytes([0x7b]),
    OP_SWAP=bytes([0x7c]),
    OP_TUCK=bytes([0x7d]),
    OP_2DROP=bytes([0x6d]),
    OP_2DUP=bytes([0x7e]),
    OP_3DUP=bytes([0x6f]),
    OP_2OVER=bytes([0x70]),
    OP_2ROT=bytes([0x71]),
    OP_2SWAP=bytes([0x72]),
))
# Splice

SPLICE = EnumDict(dict(

    OP_CAT=bytes([0x7e]),  # disabled
    OP_SUBSTR=bytes([0x7f]),  # disabled
    OP_LEFT=bytes([0x80]),  # disabled
    OP_RIGHT=bytes([0x81]),  # disabled
    OP_SIZE=bytes([0x82]),
))
# Bitwise logic
BITWISE_LOGIC = EnumDict(dict(
    OP_INVERT=bytes([0x83]),  # disabled
    OP_AND=bytes([0x84]),  # disabled
    OP_OR=bytes([0x85]),  # disabled
    OP_XOR=bytes([0x134]),  # disabled
    OP_EQUAL=bytes([0x87]),
    OP_EQUALVERIFY=bytes([0x88]),

))
# Arithmetic
ARITHMETIC = EnumDict(dict(
    OP_1ADD=bytes([0x8b]),
    OP_1SUB=bytes([0x8c]),
    OP_2MUL=bytes([0x8d]),  # disabled
    OP_2DIV=bytes([0x8e]),  # disabled
    OP_NEGATE=bytes([0x8f]),
    OP_ABS=bytes([0x90]),
    OP_NOT=bytes([0x91]),
    OP_0NOTEQUAL=bytes([0x92]),
    OP_ADD=bytes([0x93]),
    OP_SUB=bytes([0x94]),
    OP_MUL=bytes([0x95]),  # disabled
    OP_DIV=bytes([0x96]),  # disabled
    OP_MOD=bytes([0x97]),  # disabled
    OP_LSHIFT=bytes([0x98]),  # disabled
    OP_RSHIFT=bytes([0x99]),   # disabled
    OP_BOOLAND=bytes([0x9a]),
    OP_BOOLOR=bytes([0x9b]),
    OP_NUMEQUAL=bytes([0x9c]),
    OP_NUMEQUALVERIFY=bytes([0x9d]),
    OP_NUMNOTEQUAL=bytes([0x9e]),
    OP_LESSTHAN=bytes([0x9f]),
    OP_GREATERTHAN=bytes([0xa0]),
    OP_LESSTHANOREQUAL=bytes([0xa1]),
    OP_GREATERTHANOREQUAL=bytes([0xa2]),
    OP_MIN=bytes([0xa3]),
    OP_MAX=bytes([0xa4]),
    OP_WITHIN=bytes([0xa5]),
))
# Crypto
CROPTO = EnumDict(dict(
    OP_RIPEMD160=bytes([0xa6]),
    OP_SHA1=bytes([0xa7]),
    OP_SHA256=bytes([0xa8]),
    OP_HASH160=bytes([0xa9]),
    OP_HASH256=bytes([0xaa]),
    OP_CODESEPARATOR=bytes([0xab]),
    OP_CHECKSIG=bytes([0xac]),
    OP_CHECKSIGVERIFY=bytes([0xad]),
    OP_CHECKMULTISIG=bytes([0xae]),
    OP_CHECKMULTISIGVERIFY=bytes([0xaf]),
))

# Locktime

LOCKTIME = EnumDict(dict(
    OP_CHECKLOCKTIMEVERIFY=bytes([0xb1]),
    OP_CHECKSEQUENCEVERIFY=bytes([0xb2]),
))

# Pesu-words
PESU_WORDS = EnumDict(dict(
# These words are used internally for assisting with transaction matching. They are invalid if used in actual scripts.  # noqa
    OP_PUBKEYHASH=bytes([0xfd]),
    OP_PUBKEY=bytes([0xfe]),
    OP_INVALIDOPCODE=bytes([0xff]),
))
# Reserved words
RESERVED_WORDS = EnumDict(dict(
    # Any opcode not assigned is also reserved. Using an unassigned opcode makes the transaction invalid. # no qa
    OP_PERSERVED=bytes([0x50]),
    OP_VER=bytes([0x62]),
    OP_VERIF=bytes([0x65]),
    OP_VERNOTIF=bytes([0x66]),
    OP_RESERVED1=bytes([0x89]),
    OP_RESERVED2=bytes([0x8a]),

    OP_NOP1=bytes([0xb0]),
    OP_NOP4=bytes([0xb3]),
    OP_NOP5=bytes([0xb4]),
    OP_NOP6=bytes([0xb5]),
    OP_NOP7=bytes([0xb6]),
    OP_NOP8=bytes([0xb7]),
    OP_NOP9=bytes([0xb8]),
    OP_NOP10=bytes([0xb9]),
))


OPS = reduce(add, [
    RESERVED_WORDS,
    PESU_WORDS,
    LOCKTIME,
    CROPTO,
    ARITHMETIC,
    BITWISE_LOGIC,
    STACK,
    FLOW_CONTROL,
    CONSTS,
    SPLICE
])
